/*****************************************************************************
 *
 * perfparse.c - Parse Nagios data and store in relational database.
 *
 * Copyright (c) 2004 Ben Clewett (perfparse@clewett.org.uk)
 * Last Modified:   2004-05-28
 *
 * License:
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <signal.h>
#include <time.h>
#include <glib.h>
#include <sys/stat.h>
#include <libpp_common.h>
#include "db_macro_actions.h"
#include "common.h"
#include "dbms.h"
#include "convert.h"
#include <config.h>
#include <gettext.h>
#include <locale.h>

/* Debug / Output definitions: */
#undef SHOW_SQL
#undef DEBUG

#define IN_SIZE 1024

/* Domain to use: */
#define SERVICE "service"

/* Internal functions */
void print_sql_failure(char *file, int num, const char *sql, const char *error);


void my_on_exit();

/* Options */
typedef enum {
	DBTOOL_UNSET=0,
	DBTOOL_UPDATE,
	DBTOOL_INSTALL,
	DBTOOL_RESET_HOST_DELETE_POLICY
} dbtool_action_e;

dbtool_action_e dbtool_action;
char*reset_host_delete_policy_host_name;

/***************************************
 *
 * Main
 *
 ***************************************/



int main(int argc, char **argv)
{
	int options[] = {
		CONFIG_ITEM_ID_DB_DATABASE,
		CONFIG_ITEM_ID_DB_HOST,
		CONFIG_ITEM_ID_DB_PWD,
		CONFIG_ITEM_ID_DB_USER,
		CONFIG_ITEM_ID_DO_REPORT,
		CONFIG_ITEM_ID_ERROR_LOG,
		CONFIG_ITEM_ID_ERROR_LOG_ROTATE,
		CONFIG_ITEM_ID_DBTOOL_UPDATE,
		CONFIG_ITEM_ID_DBTOOL_INSTALL,
		CONFIG_ITEM_ID_DBTOOL_RESET_HOST_DELETE_POLICY,
		CONFIG_ITEM_ID_NO_TRANSACTIONS,
		CONFIG_ITEM_ID_NB
	};

	setlocale(LC_ALL, "");
	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);
	
	switch(config_parse_config(argc,argv,_("perfparse-db-tool [options]\n\n"),"\n",options, LOCALSTATEDIR "/perfparse-db-tool.log")) {
		case CONFIG_OPTION_HELP: config_print_help(stdout, _("perfparse-db-tool [options]\n\n"), "\n", options); exit(EXIT_SUCCESS);
		case CONFIG_OPTION_VERSION: config_print_version(stdout, _("perfparse-db-tool [options]\n\n"), "\n"); exit(EXIT_SUCCESS);
		case CONFIG_OPTION_SHOW_CONFIG: config_show_config(stdout, _("perfparse-db-tool [options]\n\n"), "\n", options); exit(EXIT_SUCCESS);
		case CONFIG_OPTION_NONE: break;
	}
	reset_host_delete_policy_host_name=NULL;

	/* Register the error options */
	#ifdef USE_DB_MYSQL
	libpp_mysql_set_sql_failure_callback(print_sql_failure);
	#elif defined USE_DB_POSTGRESQL
	libpp_postgresql_set_sql_failure_callback(print_sql_failure);
	#endif

	/* Connect to MySQL: */
	now_connect(TRUE);

	/* Register the on_exit: */
	atexit(my_on_exit);	

	/* exit if the database format is too old */
	if(config_get_value_as_boolean(CONFIG_ITEM_ID_DBTOOL_UPDATE)) {
		autoConvert();
	} else if(config_get_value_as_boolean(CONFIG_ITEM_ID_DBTOOL_INSTALL)) {
		install_database();
	} else if(NULL != (reset_host_delete_policy_host_name = config_get_value_as_string(CONFIG_ITEM_ID_DBTOOL_RESET_HOST_DELETE_POLICY))) {
		exit_if_db_version_not_correct();
		reset_host_delete_policy(reset_host_delete_policy_host_name);
	} else {
		config_print_help(stdout, _("perfparse-db-tool [options]\n\n"),_("No action specified\n"),options);
		exit(EXIT_FAILURE);
	}

	exit(EXIT_SUCCESS);
}


void my_on_exit()
{
	#ifdef USE_DB_MYSQL
	mysql_close(&db_mysql);
	#elif defined USE_DB_POSTGRESQL
	PQfinish(db_postgresql);
	#endif
}



void print_sql_failure(char*file,int num, const char *sql, const char *error)
{

        printf(_("An error occured with the SQL:\n\n  \"%s\"\n\nFailure Message:\n\n  \"%s\"\n\n"),
                sql, error);

        exit(EXIT_FAILURE);

}


