/*****************************************************************************
 *
 * storage_mysql.c - Parse Nagios data and store in relational database.
 *
 * Copyright (c) 2004 Ben Clewett (bclewett@perfparse.org)
 * Last Modified:   2004-09-01
 *
 * License:
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <glib.h>
#include <string.h>
#include <libnagios_perfdata_parser.h>
#include <libpp_common.h>
#include "libpp_storage.h"
#include "db_macro_actions.h"
#include "common.h"
#include "dbms.h"

#include "storage_mysql.h"
#include <config.h>
#include <gettext.h>
#include <locale.h>

#define DEBUG
#undef DEBUG

/* Binary Trees */
GTree *validated_hosts;
GTree *validated_services;
GTree *validated_metrics;
GTree *validated_summary;

/* List of summary sample values */
typedef struct TSummaryList {
	int iFrequency;
	int iDefaultDelete;
	void *next;
} TSummaryList;
TSummaryList *summaryHead = NULL;

/* Internal definitions */
int storage_mysql_connect(struct storage_module_t*);
void storage_mysql_atexit(storage_module_t*);
int  storage_mysql_store_line(storage_module_t *, const log_line_t *);
int  storage_mysql_store_perf_values(storage_module_t *, const log_line_t *, const perf_fields_t *, const perf_value_t *);
int  storage_mysql_set_registry(storage_module_t *, const char*, const char*, const char*, time_t);
void storage_mysql_print_statistics(storage_module_t *, FILE*);
void storage_mysql_print_usage(storage_module_t *this, FILE*fd, const char*id);

static int validate_host(storage_module_t*this, const char *);
static int validate_service(storage_module_t*this, const char *, const char *);
static int validate_metric(storage_module_t*this, const char *sHost, const char *sService, const char *sMetric, const char *sUnit);
static int isSaveBinAllowed(char *sHost, char *sService, char *sMetric);
static const char *getSafeD(double d);

static int save_raw_data(storage_module_t*this, const log_line_t *);
void save_bin_data(storage_module_t*this, const log_line_t *data, const perf_value_t *perf, const perf_fields_t *perf_fields, int iHostID, int iServiceID, int iMetricID);

/* Summary Data Internal Functions */
void populateSummaryList();
static int validate_summary(storage_module_t *this, int iMetricID, int iFrequency, int iDefaultDelete);
static int save_summary_data(storage_module_t *this, const log_line_t *data, const perf_value_t *perf, const perf_fields_t *perf_fields, int iMetricID);
static int save_summary_data_nested(storage_module_t *this, const log_line_t *data, const perf_value_t *perf, const perf_fields_t *perf_fields, int iMetricID, int iFrequency);

/* Raw summary internal functions */
static int save_summary_raw_data(storage_module_t*this, const log_line_t *, int iServiceID);

/********************************************************
 *
 * Initialise Call
 *
 *******************************************************/


storage_module_t* storage_module_init(void) {


	storage_module_t *sm;
	sm = storage_new("mysql", STORAGE_MYSQL_VERSION);
	if(NULL == sm) return(NULL);
	sm->connect = storage_mysql_connect;
	sm->store_line = storage_mysql_store_line;
	sm->set_registry = storage_mysql_set_registry;
	sm->store_perf_values = storage_mysql_store_perf_values;
	sm->atexit = storage_mysql_atexit;
	sm->print_statistics = storage_mysql_print_statistics;
	sm->print_usage = storage_mysql_print_usage;
	sm->enabled = 1;
	sm->params = malloc(sizeof(mysql_user_params_t));
	return(sm);
}

int storage_mysql_connect(struct storage_module_t *this) {
	if(NULL == this->params) {pp_log(__FILE__,__LINE__,_("Malloc failed")); exit(EXIT_FAILURE); }
	MYSQL_USER_PARAMS(this->params)->statistics = malloc(sizeof(mysql_statistics_t));
	if(NULL == MYSQL_USER_PARAMS(this->params)->statistics) {pp_log(__FILE__,__LINE__,_("Malloc failed")); exit(EXIT_FAILURE); }
	MYSQL_USER_PARAMS(this->params)->statistics->iAddedHost = 0;
	MYSQL_USER_PARAMS(this->params)->statistics->iAddedService = 0;
	MYSQL_USER_PARAMS(this->params)->statistics->iAddedMetric = 0;
	MYSQL_USER_PARAMS(this->params)->statistics->iAddedBinary = 0;
	MYSQL_USER_PARAMS(this->params)->statistics->iAddedRaw = 0;
	MYSQL_USER_PARAMS(this->params)->statistics->iRead = 0;
	MYSQL_USER_PARAMS(this->params)->statistics->iSQLCount = 0;
	MYSQL_USER_PARAMS(this->params)->statistics->iAddedSummaryMetric = 0;
	MYSQL_USER_PARAMS(this->params)->statistics->iAddedSummary = 0;

	/* Open MySQL */
	now_connect(TRUE);
#ifdef DEBUG
	printf("MySQL: %s\n", query_char("SELECT VERSION()"));
#endif

	/* Setup binary trees */
	validated_hosts = g_tree_new((GCompareFunc)strcmp);
	validated_services = g_tree_new((GCompareFunc)strcmp);
	validated_metrics = g_tree_new((GCompareFunc)strcmp);
	validated_summary = g_tree_new((GCompareFunc)strcmp);

	/* Get list of summary tables */
	populateSummaryList();

	/* exit if the database format is too old */
	if (is_db_incorrect_version()) {
		gchar *str = db_version_error();
		this->enabled = 0;
		pp_log(__FILE__, __LINE__, "%s", str);
		g_free(str);
		exit(EXIT_FAILURE);
	}


	/* Switch on full transactions */
	if (!config_get_value_as_boolean(CONFIG_ITEM_ID_NO_TRANSACTIONS)) {
		query("SET AUTOCOMMIT=0");
	}


	MYSQL_USER_PARAMS(this->params)->statistics->tStart = time(NULL);

	/* Ready for lines of data */

	return(0);
}


/********************************************************
 *
 * Registered calls
 *
 *******************************************************/

int storage_mysql_store_line(storage_module_t *this, const log_line_t *log_data) {

	int iHostID, iServiceID;

#ifdef DEBUG
	printf("MY(store_line) macro[LOG_LINE_TIMET]        = '%s'\n", log_data->macro[LOG_LINE_TIMET]);
	printf("MY(store_line) macro[LOG_LINE_HOSTNAME]     = '%s'\n", log_data->macro[LOG_LINE_HOSTNAME]);
	printf("MY(store_line) macro[LOG_LINE_SERVICEDESC]  = '%s'\n", log_data->macro[LOG_LINE_SERVICEDESC]);
	printf("MY(store_line) macro[LOG_LINE_OUTPUT]       = '%s'\n", log_data->macro[LOG_LINE_OUTPUT]);
	printf("MY(store_line) macro[LOG_LINE_SERVICESTATE] = '%s'\n", log_data->macro[LOG_LINE_SERVICESTATE]);
	fflush(stdout);
#endif

	/* Start transaction */
	if (!config_get_value_as_boolean(CONFIG_ITEM_ID_NO_TRANSACTIONS)) {
		query("START TRANSACTION");
	}

	/* Validate Host */
	iHostID = validate_host(this, log_data->macro[LOG_LINE_HOSTNAME]);

	/* Validate service */
	iServiceID = validate_service(this, log_data->macro[LOG_LINE_HOSTNAME], log_data->macro[LOG_LINE_SERVICEDESC]);

	/* Save the raw data */
	save_raw_data(this, log_data);

    /* Summarize availability data */       
    save_summary_raw_data(this, log_data, iServiceID);

	/* Commit transaction */
	if (!config_get_value_as_boolean(CONFIG_ITEM_ID_NO_TRANSACTIONS)) {
		query("COMMIT");
	}

	MYSQL_USER_PARAMS(this->params)->statistics->iRead++;

	return(0);
}






int storage_mysql_store_perf_values(storage_module_t *this, const log_line_t *log_data, const perf_fields_t *perf_fields, const perf_value_t *perf_value) {

	int iHostID, iServiceID, iMetricID;

#ifdef DEBUG
	setlocale(LC_NUMERIC, "POSIX"); 
	if((NULL != perf_value) && (perf_fields->type == PERF_FIELDS_TYPE_PERF)) {
		printf("MY(store_perf_values) %s => {\n",				perf_fields->label);
		printf("MY(store_perf_values) \t\thostname      = '%s',\n",	log_data->macro[LOG_LINE_HOSTNAME]);
		printf("MY(store_perf_values) \t\tservice       = '%s',\n",	log_data->macro[LOG_LINE_SERVICEDESC]);
		printf("MY(store_perf_values) \t\tmetric        = '%s',\n",	perf_fields->label);
		printf("MY(store_perf_values) \t\tvalue         = %.10e,\n",	perf_value->d[PERF_VALUE_VALUE]);
		printf("MY(store_perf_values) \t\tunit          = '%s',\n",	perf_value->unit);
		printf("MY(store_perf_values) \t\twarn_start    = %.10e,\n",	perf_value->d[PERF_VALUE_WARN_START]);
		printf("MY(store_perf_values) \t\twarn_end      = %.10e,\n",	perf_value->d[PERF_VALUE_WARN_END]);
		printf("MY(store_perf_values) \t\twarn_inverted = %s,\n",	(perf_value->ranges_type & PERF_VALUE_RANGE_WARNING_RANGE_INSIDE)?"yes":"no");
		printf("MY(store_perf_values) \t\tcrit_start    = %.10e,\n",	perf_value->d[PERF_VALUE_CRIT_START]);
		printf("MY(store_perf_values) \t\tcrit_end      = %.10e,\n",	perf_value->d[PERF_VALUE_CRIT_END]);
		printf("MY(store_perf_values) \t\tcrit_inverted = %s,\n",	(perf_value->ranges_type & PERF_VALUE_RANGE_CRITICAL_RANGE_INSIDE)?"yes":"no");
		printf("MY(store_perf_values) \t\t\t}\n");
		fflush(stdout);
	}
	setlocale(LC_NUMERIC, ""); 
#endif

	/* Start transaction */
	if (!config_get_value_as_boolean(CONFIG_ITEM_ID_NO_TRANSACTIONS)) {
		query("START TRANSACTION");
	}

	/* Validate Host */
	iHostID = validate_host(this, log_data->macro[LOG_LINE_HOSTNAME]);

	/* Validate service */
	iServiceID = validate_service(this, log_data->macro[LOG_LINE_HOSTNAME], log_data->macro[LOG_LINE_SERVICEDESC]);

	/* Validate Metric */
	iMetricID = validate_metric(this, log_data->macro[LOG_LINE_HOSTNAME], log_data->macro[LOG_LINE_SERVICEDESC], perf_fields->label, perf_value->unit);

	/* Check policy */
	/* [TODO] use a hash-table for efficient re-checking */
	if (!isSaveBinAllowed(log_data->macro[LOG_LINE_HOSTNAME], log_data->macro[LOG_LINE_SERVICEDESC], perf_fields->label)) return(0);

	/* Save the binary data */
	save_bin_data(this, log_data, perf_value, perf_fields, iHostID, iServiceID, iMetricID);

	/* Save the Summary Data, if any */
	save_summary_data(this, log_data, perf_value, perf_fields, iMetricID);


	/* Commit transaction */
	if (!config_get_value_as_boolean(CONFIG_ITEM_ID_NO_TRANSACTIONS)) {
		query("COMMIT");
	}

	return(0);
}



int storage_mysql_set_registry(storage_module_t *this, const char*hostname, const char*key, const char*value, time_t tm) {

#ifdef DEBUG_REGISTRY
	printf("MY(set_registry) %ld '%s/%s' = '%s'\n", tm,hostname,key,value);
#endif

	setRegistry(hostname, key, value, tm);

	return(0);
}


void storage_mysql_atexit(storage_module_t* this)
{

	mysql_close(&db_mysql);

}

void storage_mysql_print_statistics(storage_module_t *this, FILE*fd) {

	MYSQL_USER_PARAMS(this->params)->statistics->iSQLCount = iSQLCount;

	if (!config_get_value_as_boolean(CONFIG_ITEM_ID_DO_REPORT)) return;

	fprintf(fd, "+-----------------------------+   +-----------------------------+\n");
	fprintf(fd, "| MySQL storage statistics    |   |                             |\n");
	fprintf(fd, "|   New Hosts :       %7d |   | Metric/line :       %7.2f |\n", 
		MYSQL_USER_PARAMS(this->params)->statistics->iAddedHost,
		(MYSQL_USER_PARAMS(this->params)->statistics->iRead) ? 
		(double)MYSQL_USER_PARAMS(this->params)->statistics->iAddedBinary / 
		(double)MYSQL_USER_PARAMS(this->params)->statistics->iRead : (double)0.);
	fprintf(fd, "|   New Metrics :     %7d |   | Metrics Recorded :  %7d |\n", 
		MYSQL_USER_PARAMS(this->params)->statistics->iAddedMetric,
		MYSQL_USER_PARAMS(this->params)->statistics->iAddedBinary);
	fprintf(fd, "|   New Services :    %7d |   | Summary Rec Added : %7d |\n", 
		MYSQL_USER_PARAMS(this->params)->statistics->iAddedService,
		MYSQL_USER_PARAMS(this->params)->statistics->iAddedSummary);
	fprintf(fd, "|   New Summary Group : %5d |   | SQL Queries :       %7d |\n", 
		MYSQL_USER_PARAMS(this->params)->statistics->iAddedSummaryMetric,
		MYSQL_USER_PARAMS(this->params)->statistics->iSQLCount);
	fprintf(fd, "+-----------------------------+   +-----------------------------+\n");

}

void storage_mysql_print_usage(storage_module_t *this, FILE*fd, const char*id) {
	int options[] = {
		CONFIG_ITEM_ID_DB_USER,
		CONFIG_ITEM_ID_DB_PWD,
		CONFIG_ITEM_ID_DB_DATABASE,
		CONFIG_ITEM_ID_DB_HOST,
		CONFIG_ITEM_ID_NO_TRANSACTIONS,
		CONFIG_ITEM_ID_NB
	};
	if(!strcmp(id,"version")) {
		fprintf(fd, _("Storage Module : %s\nVersion: %s (built %s %s)\n"),
				this->id, this->version, __DATE__, __TIME__);
	} else if(!strcmp(id, "help")) {
		fprintf(fd, _("Storage Module : %s\n"), this->id);
		config_print_help(fd, NULL,"", options);
	} else if(!strcmp(id, "show_config")) {
		fprintf(fd, _("# Storage Module : %s\n"), this->id);
		fprintf(fd, _("# ==============================\n\n"));
		config_show_config(fd, NULL,"", options);
		fprintf(fd, "\n");
	}
}

/********************************************************
 *
 * Internal Functions
 *
 *******************************************************/


/********************************************************
 *
 * Validation
 *
 *******************************************************/

static int validate_host(storage_module_t*this, const char *sHost)
{
	/* Validate the host exists.  If not, create */

	static int *iHostID, iTestHost;
	static char *sKey;

	/* Have we seen this one before? */
	iHostID = g_tree_lookup(validated_hosts, sHost);
	if (iHostID) return *iHostID;

	/* Does host exist ? */
	g_string_printf(s_SQL, "SELECT host_id FROM perfdata_host WHERE host_name = '%s'",
		sql_escape(sHost,PERCENT_ESCAPE_CHAR));
	iTestHost = query_int(s_SQL->str);

	if (!iTestHost) {

		/* New host.  Add to database. */
		iTestHost = query_int("SELECT MAX(host_id) FROM perfdata_host");
		iTestHost++;

		g_string_printf(s_SQL, "INSERT INTO perfdata_host (host_name, host_id) VALUES ('%s', %d)", 
			sql_escape(sHost,PERCENT_ESCAPE_CHAR), iTestHost);
		query(s_SQL->str);

		MYSQL_USER_PARAMS(this->params)->statistics->iAddedHost++;

	}

	/* Add to bitree */
	iHostID = (int *)malloc(sizeof(int));
	*iHostID = iTestHost;
	sKey = strdup(sHost);
	g_tree_insert(validated_hosts, sKey, iHostID);

	return *iHostID;


}



static int validate_service(storage_module_t*this, const char *sHost, const char *sService)
{

	/* Validate the service exists.  If not, create */

	static char sHash[200];
	static char *sKey;
	static int *iServiceID, iTestService;


	/* Have we seen this one before? */
	sprintf(sHash, "%s:%s", sService, sHost);
	iServiceID = g_tree_lookup(validated_services, sHash);
	if (iServiceID) return *iServiceID;


	/* Does service exist ? */
	g_string_printf(s_SQL, "SELECT service_id FROM perfdata_service WHERE host_name = '%s'", 
		sql_escape(sHost,PERCENT_ESCAPE_CHAR));
	g_string_append_printf(s_SQL, " AND service_description = '%s'",
		sql_escape(sService,PERCENT_ESCAPE_CHAR));
	iTestService = query_int(s_SQL->str);

	if (!iTestService) {

		/* New service.  Add to database. */
		iTestService = query_int("SELECT MAX(service_id) FROM perfdata_service");
		iTestService++;

		g_string_printf(s_SQL, "INSERT INTO perfdata_service (host_name, service_description, raw_delete_policy_type, service_id) VALUES ('%s'",
			sql_escape(sHost,PERCENT_ESCAPE_CHAR));
		g_string_append_printf(s_SQL, ", '%s', 'host', %d)",
			sql_escape(sService,PERCENT_ESCAPE_CHAR), iTestService);
		query(s_SQL->str);

		MYSQL_USER_PARAMS(this->params)->statistics->iAddedService++;

	}

	/* Does service_raw entry exist? */
    g_string_printf(s_SQL, "SELECT COUNT(*) FROM perfdata_raw_summary WHERE service_id = %d", iTestService);
	if (!query_int(s_SQL->str)) {
		/* new summary_raw entry */
        /* insert a default entry for this service */
        /* TODO: make the default epoch (1 day) configurable? */
        g_string_printf(s_SQL, "INSERT INTO perfdata_raw_summary (service_id, epoch) VALUES (%d, %d)", iTestService, 86400);
        query(s_SQL->str);
	}


	/* Add to bitree */
	iServiceID = (int *)malloc(sizeof(int));
	*iServiceID = iTestService;
	sKey = strdup(sHash);
	g_tree_insert(validated_services, sKey, iServiceID);

	return *iServiceID;

}




static int validate_metric(storage_module_t*this, const char *sHost, const char *sService, const char *sMetric, const char *sUnit)
{
	/* Validate the metric exists.  If not, create */

	static int *iMetricID, iTestMetric;
	static char sHash[200];
	static char *sKey;

	/* Have we seen this one before? */
	sprintf(sHash, "%100s:%45s:%45s", sMetric, sService, sHost);
	iMetricID = g_tree_lookup(validated_services, sHash);
	if (iMetricID) return *iMetricID;

	/* Does metric exist ? */
	g_string_printf(s_SQL, "SELECT metric_id FROM perfdata_service_metric WHERE host_name = '%s'",
		sql_escape(sHost,PERCENT_ESCAPE_CHAR));
	g_string_append_printf(s_SQL, " AND service_description = '%s'",
		sql_escape(sService,PERCENT_ESCAPE_CHAR));
	g_string_append_printf(s_SQL, " AND metric = '%s'", 
		sql_escape(sMetric,PERCENT_ESCAPE_CHAR));
	query(s_SQL->str);
	iTestMetric = query_int(s_SQL->str);

	if (!iTestMetric) {

		/* New metric.  Add to database */
		iTestMetric = query_int("SELECT MAX(metric_id) FROM perfdata_service_metric");
		iTestMetric++;

		g_string_printf(s_SQL, "INSERT INTO perfdata_service_metric (host_name, service_description, metric, unit, bin_delete_policy_type, metric_id) VALUES ('%s'",
			sql_escape(sHost,PERCENT_ESCAPE_CHAR));
		g_string_append_printf(s_SQL, ", '%s'",
			sql_escape(sService,PERCENT_ESCAPE_CHAR));
		g_string_append_printf(s_SQL, ", '%s'",
			sql_escape(sMetric,PERCENT_ESCAPE_CHAR));
		g_string_append_printf(s_SQL, ", '%s', 'host', %d)",
			sql_escape(sUnit,PERCENT_ESCAPE_CHAR), iTestMetric);
		query(s_SQL->str);

		MYSQL_USER_PARAMS(this->params)->statistics->iAddedMetric++;

	}


	/* Add to bitree */
	iMetricID = (int *)malloc(sizeof(int));
	*iMetricID = iTestMetric;
	sKey = g_strdup(sHash);
	g_tree_insert(validated_metrics, sKey, iMetricID);

	return *iMetricID;


}


static int isSaveBinAllowed(char *sHost, char *sService, char *sMetric)
{
	/* Check policy to find out if we are allowed to save this metric */
	static int bResult;

	g_string_printf(s_SQL, "SELECT bin_delete_policy FROM perfdata_service_metric WHERE host_name = '%s'",
		sql_escape(sHost, PERCENT_ESCAPE_CHAR)); 
	g_string_append_printf(s_SQL, " AND service_description = '%s'",
		sql_escape(sService, PERCENT_ESCAPE_CHAR));
	g_string_append_printf(s_SQL, " AND metric = '%s'",
		sql_escape(sMetric, PERCENT_ESCAPE_CHAR));
	query(s_SQL->str);
	result_row = mysql_fetch_row(query_result);

	if (rows() == 0) return TRUE;

	if (!result_row[0] || iData(0) > 0) 
		bResult = TRUE;
	else
		bResult = FALSE;

	return bResult;


}



/********************************************************
 *
 * Save Data
 *
 *******************************************************/

static int save_raw_data(storage_module_t*this, const log_line_t *data)
{
	static int iAuto;

	g_string_printf(s_SQL, "INSERT INTO perfdata_service_raw (host_name, service_description, ctime, txt_data, nagios_status) VALUES ('%s'",
		sql_escape(data->macro[LOG_LINE_HOSTNAME], PERCENT_ESCAPE_CHAR));
	g_string_append_printf(s_SQL, ",'%s',FROM_UNIXTIME(%ld)",
		 sql_escape(data->macro[LOG_LINE_SERVICEDESC], PERCENT_ESCAPE_CHAR), data->time);
	g_string_append_printf(s_SQL, ",'%s'",
		sql_escape(data->macro[LOG_LINE_OUTPUT], PERCENT_ESCAPE_CHAR));
	g_string_append_printf(s_SQL, ",%d)", data->service_state);
	query(s_SQL->str);

	/* Get the auto-increment */
	query("SELECT LAST_INSERT_ID()");
	result_row = mysql_fetch_row(query_result);
	iAuto = iData(0);

	g_string_printf(s_SQL, "UPDATE perfdata_service SET last_perfdata_raw = %d WHERE host_name = '%s'",
		iAuto, sql_escape(data->macro[LOG_LINE_HOSTNAME], PERCENT_ESCAPE_CHAR)); 
	g_string_append_printf(s_SQL, " AND service_description = '%s'",
		sql_escape(data->macro[LOG_LINE_SERVICEDESC], PERCENT_ESCAPE_CHAR));
	query(s_SQL->str);

	MYSQL_USER_PARAMS(this->params)->statistics->iAddedRaw++;

	return TRUE;
}

static int save_summary_raw_data(storage_module_t*this, const log_line_t *data, int iServiceID)
{
        /* first update any existing records
        only do this if the last update time is smaller than ctime
        */

        g_string_sprintf(s_SQL, "update perfdata_raw_summary_data D JOIN perfdata_raw_summary R "
                "ON D.service_id = R.service_id AND D.epoch = R.epoch "
                "set "
                "D.sum_uptime =       IF(R.last_nagios_status = 0,D.sum_uptime + %ld - R.last_ctime,  D.sum_uptime), "
                "D.sum_warntime =     IF(R.last_nagios_status = 1,D.sum_warntime + %ld - R.last_ctime,  D.sum_warntime), "
                "D.sum_criticaltime = IF(R.last_nagios_status = 2,D.sum_criticaltime + %ld - R.last_ctime,  D.sum_criticaltime), "
                "D.sum_undeftime =    IF(R.last_nagios_status = 3,D.sum_undeftime + %ld - R.last_ctime,  D.sum_undeftime), "
                "D.count =            D.count + 1 "
                "where D.service_id = %d AND R.last_ctime < %ld AND D.epoch_start = %ld - MOD(%ld, R.epoch)",
				data->time, data->time, data->time, data->time,
				iServiceID, data->time, data->time, data->time);

        query(s_SQL->str);

        /* insert new entries for those where a new epoch has started
           do this by  just inserting new records for all using
           INSERT IGNORE, and thus making use of the primary key of the
           table so that only new entries get inserted */
        g_string_sprintf(s_SQL, "insert IGNORE into perfdata_raw_summary_data "
                        "(service_id , epoch_start , epoch, count, "
                        "sum_uptime , sum_warntime , sum_criticaltime , sum_undeftime ) "
                        "select service_id, %ld - MOD(%ld, R.epoch),  R.epoch, 1, "
                        "IF(R.last_ctime > 0 AND R.last_nagios_status = 0,  %ld - R.last_ctime,0), "
                        "IF(R.last_ctime > 0 AND R.last_nagios_status = 1,  %ld - R.last_ctime,0), "
                        "IF(R.last_ctime > 0 AND R.last_nagios_status = 2,  %ld - R.last_ctime,0), "
                        "IF(R.last_ctime > 0 AND R.last_nagios_status = 3,  %ld - R.last_ctime,0) "
                        "from perfdata_raw_summary R where service_id=%d",
						data->time, data->time, data->time, data->time, data->time, data->time,
						iServiceID);

        query(s_SQL->str);

        /* update the _raw table */
        g_string_sprintf(s_SQL, "update perfdata_raw_summary set "
                        "last_ctime = %ld , last_nagios_status = %d "
                        "where service_id=%d",
						data->time, data->service_state, iServiceID);

        query(s_SQL->str);

        return TRUE;
}

/* Save the binary data */
void save_bin_data(storage_module_t*this, const log_line_t *data, const perf_value_t *perf, const perf_fields_t *perf_fields, int iHostID, int iServiceID, int iMetricID)
{
	int iAuto = 0;

	/* Save record */
	g_string_printf(s_SQL, "INSERT INTO perfdata_service_bin (host_name, service_description, metric, ctime, value, warn, critical, state) VALUES ('%s'",
		sql_escape(data->macro[LOG_LINE_HOSTNAME], PERCENT_ESCAPE_CHAR));
	g_string_append_printf(s_SQL, ", '%s'",
		 sql_escape(data->macro[LOG_LINE_SERVICEDESC] ,PERCENT_ESCAPE_CHAR));
	g_string_append_printf(s_SQL, ", '%s', FROM_UNIXTIME(%ld), %s",
		sql_escape(perf_fields->label, PERCENT_ESCAPE_CHAR), data->time,
		getSafeD(perf->d[PERF_VALUE_VALUE]));
	g_string_append_printf(s_SQL, ", %s",
		getSafeD(perf->d[PERF_VALUE_WARN_END]));
	g_string_append_printf(s_SQL, ", %s",
		getSafeD(perf->d[PERF_VALUE_CRIT_END]));
	g_string_append_printf(s_SQL, ", %s)",
		getSafeD(perf->metric_state));

#ifdef DEBUG
	printf("\"%s\"\n", s_SQL->str);
	printf("WARN: %s\n"
			"CRIT: %s\n",
			getSafeD(perf->d[PERF_VALUE_WARN_START]),
			getSafeD(perf->d[PERF_VALUE_CRIT_START]));
#endif

	query(s_SQL->str);

	/* Get the auto-increment */
	query("SELECT LAST_INSERT_ID()");
	result_row = mysql_fetch_row(query_result);
	iAuto = iData(0);

	g_string_printf(s_SQL, "UPDATE perfdata_service_metric SET last_perfdata_bin = %d", iAuto);

	if (perf->d[PERF_VALUE_MIN] != 0 || perf->d[PERF_VALUE_MAX] != 0) {
		g_string_append_printf(s_SQL, ", value_min = %s", getSafeD(perf->d[PERF_VALUE_MIN]));
		g_string_append_printf(s_SQL, ", value_max = %s", getSafeD(perf->d[PERF_VALUE_MAX]));
	}
	
	g_string_append_printf(s_SQL, " WHERE host_name = '%s'",
		sql_escape(data->macro[LOG_LINE_HOSTNAME], PERCENT_ESCAPE_CHAR));
	g_string_append_printf(s_SQL, " AND service_description = '%s'",
		sql_escape(data->macro[LOG_LINE_SERVICEDESC] ,PERCENT_ESCAPE_CHAR));
	g_string_append_printf(s_SQL, " AND metric = '%s'",
		sql_escape(perf_fields->label, PERCENT_ESCAPE_CHAR));
	query(s_SQL->str);

	MYSQL_USER_PARAMS(this->params)->statistics->iAddedBinary++;

}


static const char *getSafeD(double d)
{
	static char sOut[25];

	setlocale(LC_NUMERIC, "POSIX"); 

	if (isnan(d)) {
		strcpy(sOut, "NULL");
	} else if (isinf(d)) {
		strcpy(sOut, "NULL");
	} else {
  		sprintf(sOut, "%.14e", d);
	}

	setlocale(LC_NUMERIC, ""); 

	return sOut;

}




/*
*
 * Functions concerened with the storage of Summary Data
 *
 */


/* Summary Data Internal Functions */
void populateSummaryList()
{
	TSummaryList *summary, *tail = NULL;

	g_string_printf(s_SQL, "SELECT frequency, default_delete FROM perfdata_bin_summary WHERE is_deleted = 0");
	query(s_SQL->str);

    while ((result_row = mysql_fetch_row(query_result))) {

		summary = (TSummaryList *)malloc(sizeof(TSummaryList));

		summary->iFrequency = iData(0);
		summary->iDefaultDelete = iData(1);
		summary->next = NULL;

		if (tail) {
			tail->next = summary;
		} else {
			summaryHead = summary;
		}

		tail = summary;

	}



}


static int save_summary_data(storage_module_t *this, const log_line_t *data, const perf_value_t *perf, const perf_fields_t *perf_fields, int iMetricID)
{
	TSummaryList *summary;

	/* Itterate through frequencies */
	summary = summaryHead;

	while (summary) {

		/* Do we store data? */
		if (validate_summary(this, iMetricID, summary->iFrequency, summary->iDefaultDelete)) {
			/* Store */
			save_summary_data_nested(this, data, perf, perf_fields, iMetricID, summary->iFrequency);
		}

		summary = summary->next;
	}

	return(0);
}

static int save_summary_data_nested(storage_module_t *this, const log_line_t *data, const perf_value_t *perf, const perf_fields_t *perf_fields, int iMetricID, int iFrequency)
{

	time_t tSample;
	int iCount;
	double dVal, dSum, dSumSquare, dMax, dMin;

	dVal = perf->d[PERF_VALUE_VALUE];


	/* Time of sample.  (Frequency is in hours.) */
	tSample = data->time - (data->time % (iFrequency * 60 * 60));

	/* Get existing record */
	g_string_printf(s_SQL, "SELECT val_count, sum_val, sum_square_val, max_val, min_val FROM perfdata_bin_summary_data");
	g_string_append_printf(s_SQL, " WHERE metric_id = %d AND frequency = %d AND ctime = FROM_UNIXTIME(%ld)",
		iMetricID, iFrequency, (long)tSample);
	query(s_SQL->str);

    result_row = mysql_fetch_row(query_result);

	if (result_row) {

		/* Exists. */
		iCount = iData(0) + 1;
		dSum = atof(sData(1)) + dVal;
		dSumSquare = atof(sData(2)) + pow(dVal, 2);
		dMax = atof(sData(3));
		dMin = atof(sData(4));
		if (dVal > dMax) dMax = dVal;
		if (dVal < dMin) dMin = dVal;

		g_string_printf(s_SQL, "UPDATE perfdata_bin_summary_data SET val_count = %d, sum_val = %s",
			iCount, getSafeD(dSum));
		g_string_append_printf(s_SQL, ", sum_square_val = %s", getSafeD(dSumSquare));
		g_string_append_printf(s_SQL, ", max_val = %s", getSafeD(dMax));
		g_string_append_printf(s_SQL, ", min_val = %s", getSafeD(dMin));
		g_string_append_printf(s_SQL, " WHERE metric_id = %d AND frequency = %d AND ctime = FROM_UNIXTIME(%ld)",
			iMetricID, iFrequency, (long)tSample);
		query_no_return(s_SQL->str);

	}

	else {

		/* New record */
		iCount = 1;
		dSum = dVal;
		dSumSquare = pow(dVal, 2);
		dMax = dVal;
		dMin = dVal;

		g_string_printf(s_SQL, "INSERT perfdata_bin_summary_data (metric_id, frequency, ctime, val_count, sum_val, sum_square_val, max_val, min_val) VALUES (%d, %d, FROM_UNIXTIME(%ld), %d, %s",
			iMetricID, iFrequency, (long)tSample, iCount, getSafeD(dSum));
		g_string_append_printf(s_SQL, ", %s", getSafeD(dSumSquare));
		g_string_append_printf(s_SQL, ", %s", getSafeD(dMax));
		g_string_append_printf(s_SQL, ", %s)", getSafeD(dMin));
		query_no_return(s_SQL->str);

	}

	MYSQL_USER_PARAMS(this->params)->statistics->iAddedSummary++;

	/* That's it... */

	return(0);
}


static int validate_summary(storage_module_t *this, int iMetricID, int iFrequency, int iDefaultDelete)
{

	/* Validate the summary data is correct. */
	/* May have to create a record */

	/* Return:  TRUE = use, FALSE = don't use. */

	static char sHash[200];
	static char *sKey;
	static int *iUse;

	/* Hash Code */
	sprintf(sHash, "%d:%d", iMetricID, iFrequency);

	/* Has this one been validated? */
	iUse = g_tree_lookup(validated_summary, sHash);
	if (iUse) return *iUse;

	/* To create new bitree record: */
	iUse = (int *)malloc(sizeof(int));

	/* Does exist? */
	g_string_printf(s_SQL, "SELECT bin_delete_policy FROM perfdata_bin_summary_header");
	g_string_append_printf(s_SQL, " WHERE metric_id = %d AND frequency = %d", iMetricID, iFrequency);
	query(s_SQL->str);

    result_row = mysql_fetch_row(query_result);

	if (result_row) {

		/* Exists. */
		if (iData(0) > 0) {
			/* Yes, use this one. */
			*iUse = TRUE;
		} else {
			/* No, don't use this one. */
			*iUse = FALSE;
		}

	} else {

		/* New record */
		g_string_printf(s_SQL, "INSERT perfdata_bin_summary_header (metric_id, frequency, last_data, bin_delete_policy, bin_delete_policy_name, bin_delete_policy_type) VALUES (%d,%d,NULL,%d,'default','template')",
			iMetricID, iFrequency, iDefaultDelete);
		query_no_return(s_SQL->str);

		MYSQL_USER_PARAMS(this->params)->statistics->iAddedSummaryMetric++;

		*iUse = TRUE;
	}

	/* Add to bitree */
	sKey = g_strdup(sHash);
	g_tree_insert(validated_summary, sKey, iUse);

	return *iUse;


}




