/*****************************************************************************
 *
 * storage_history.c - Parse Nagios data and store in relational database.
 *
 * Copyright (c) 2004 Yves Mettier (ymettier@libertysurf.fr)
 * Last Modified:   2004-10-27
 *
 * License:
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <glib.h>
#include <zlib.h>
#include <time.h>
#include <libpp_common.h>
#include "storage_history.h"
#include <libnagios_perfdata_parser.h>

#include <config.h>
#include <gettext.h>

time_t history_tm_start=0;
time_t history_tm_end=0;
time_t history_tm_cur=0;
char*history_host=NULL;
char*history_metric=NULL;
char*history_key=NULL;
gzFile*history_fd = NULL;
time_t history_fd_date=0;
log_line_t *history_log_data=NULL;
perf_fields_t*history_perf_fields=NULL;


int history_log_file_open(time_t tm) {
	gchar *path;
	time_t fd_date = tm/86400;
	char date[] = "YYYY/MM/YYYYMMDD.log";
	struct stat buf;

	if(history_fd) {
		if(fd_date == history_fd_date) return(0);
		gzclose(history_fd);
		history_fd = NULL;
	}
	if(config_get_value_as_boolean(CONFIG_ITEM_ID_STORAGE_FILE_OUTPUT_LOG_ROTATE)) {
		const struct tm *stm = localtime(&tm);
		strftime(date, sizeof(date), "%Y/%m/%Y%m%d.log",stm);
		path = g_strdup_printf("%s/%s.gz",config_get_value_as_string(CONFIG_ITEM_ID_STORAGE_FILE_OUTPUT_LOG_FILENAME),
				date);
	} else {
		path = g_strdup_printf("%s.gz",config_get_value_as_string(CONFIG_ITEM_ID_STORAGE_FILE_OUTPUT_LOG_FILENAME));
	}
	if(stat(path,&buf)) {
		path[strlen(path)-3] = '\0';
	}
	history_fd = gzopen(path,"r");
	g_free(path);
	history_fd_date = tm/86400;

	return(history_fd?0:-1);
}

GString *history_read_line(GString*line) {
	GString *l = line;
	char*lcr;
	char history_buffer[BUFSIZ];

	while(Z_NULL != (lcr = gzgets(history_fd, history_buffer, BUFSIZ))) {
		char*cr = strchr(history_buffer,'\n');
		if(cr) {
			cr[0] = '\0';
			l = g_string_append(l,history_buffer);
			break;
		}
		l = g_string_append(l,history_buffer);
	}
	return(l);
}

GString* history_read_next_line(GString*line) {
	GString*l;
	l = g_string_assign(line,"");
	while(0 == l->len) {
		while(0 != history_log_file_open(history_tm_cur)) {
			history_tm_cur += 86400;
			if(history_tm_cur > history_tm_end) {
				g_string_free(l,TRUE);
				return(NULL);
			}
		}
		l = history_read_line(l);
		if(0 == l->len) {
			gzclose(history_fd);
			history_fd = NULL;
			history_tm_cur += 86400;
			if(history_tm_cur > history_tm_end) {
				g_string_free(l,TRUE);
				return(NULL);
			}
		}
	}
	return(l);
}

GString* history_get_lines_next(GString*line) {
	GString*l = line;
	time_t tm=0;
	while((l = history_read_next_line(l))) {
		tm = strtol(l->str,NULL,10);
		if((tm >= history_tm_start) && (tm <= history_tm_end)) {
			if(history_host || history_metric || history_key) {
				if(line_split(l->str,history_log_data)) continue;
				else {
					if(history_host && strcmp(history_host, history_log_data->macro[LOG_LINE_HOSTNAME])) continue;
					if(history_metric && strcmp(history_metric, history_log_data->macro[LOG_LINE_SERVICEDESC])) continue;
					if(history_key) {
						if(perf_get_first(history_log_data,history_perf_fields)) continue;
						else {
							int notfound=1; 
							do {
								notfound = strcmp(history_key,history_perf_fields->label);
							} while(notfound && !perf_get_next(history_log_data,history_perf_fields));
							if(notfound) continue;
						}
					} else break;
				}
			}
			break;
		}
	}
	if(l && (tm > history_tm_end)) {
		g_string_free(l, TRUE);
		gzclose(history_fd);
		history_fd = NULL;
		l = NULL;

	}
	return(l);
}

GString* history_get_lines_first(time_t tm_start, time_t tm_end, const char*host, const char*metric, const char*key) {
	history_tm_start = tm_start;
	history_tm_end = tm_end;
	history_host = host[0]?strdup(host):NULL;
	history_metric = metric[0]?strdup(metric):NULL;
	history_key = key[0]?strdup(key):NULL;
	history_tm_cur = tm_start - (tm_start%86400);
	if(NULL == history_log_data) history_log_data = log_line_new();
	if(NULL == history_perf_fields) history_perf_fields = perf_fields_new();
	return(history_get_lines_next(g_string_new("")));
}


