--
-- Tables used for perfparse. PostgreSQL.
--
--
-- Structure:
--
-- 'service' and 'host' relations.  
-- To be dropped in future when replacement
-- for serive.cfg is added.
--

CREATE TABLE perfdata_delete_policy (
	policy_name VARCHAR(20) NOT NULL PRIMARY KEY,
	delete_policy INT DEFAULT NULL
);

CREATE TABLE perfdata_host_group (
	group_name VARCHAR(75) PRIMARY KEY
);

CREATE TABLE perfdata_host (
	host_id INT NOT NULL,
	--host_id SERIAL NOT NULL,
	UNIQUE(host_id),
	host_name VARCHAR(75) NOT NULL PRIMARY KEY,
	group_name VARCHAR(75),
	FOREIGN KEY (group_name) 
		REFERENCES perfdata_host_group (group_name) 
		ON DELETE SET NULL,
	raw_delete_policy INTEGER,
	raw_delete_policy_name VARCHAR(20) DEFAULT NULL,
	raw_delete_policy_type VARCHAR(10) CHECK (raw_delete_policy_type IN ('user','template')) NOT NULL, 
	FOREIGN KEY (raw_delete_policy_name)
		REFERENCES perfdata_delete_policy (policy_name)
		ON DELETE SET NULL,
	bin_delete_policy INTEGER,
	bin_delete_policy_name VARCHAR(20) DEFAULT NULL,
	bin_delete_policy_type VARCHAR(10) CHECK (bin_delete_policy_type IN ('user','template')) NOT NULL, 
	FOREIGN KEY (bin_delete_policy_name)
		REFERENCES perfdata_delete_policy (policy_name)
		ON DELETE SET NULL,
	is_deleted BOOL DEFAULT FALSE
	);
	
CREATE INDEX perfdata_host_ix0 ON perfdata_host (group_name);
CREATE INDEX perfdata_host_ix1 ON perfdata_host (raw_delete_policy_name);
CREATE INDEX perfdata_host_ix2 ON perfdata_host (bin_delete_policy_name);

CREATE TABLE perfdata_service (
	service_id INT NOT NULL,
	--service_id SERIAL NOT NULL,
	UNIQUE(service_id),
	host_name VARCHAR(75) NOT NULL,
	service_description VARCHAR(75) NOT NULL,
	PRIMARY KEY (host_name, service_description),
	FOREIGN KEY (host_name) 
		REFERENCES perfdata_host (host_name) 
		ON DELETE RESTRICT,
	last_perfdata_raw INTEGER DEFAULT NULL,
	raw_delete_policy INTEGER DEFAULT NULL,
	raw_delete_policy_name VARCHAR(20) DEFAULT NULL,
	raw_delete_policy_type VARCHAR(10) CHECK (raw_delete_policy_type IN('user','template','host')) NOT NULL, 
	FOREIGN KEY (raw_delete_policy_name)
		REFERENCES perfdata_delete_policy (policy_name)
		ON DELETE SET NULL,
	is_deleted BOOL DEFAULT FALSE
	);
	
CREATE INDEX perfdata_service_ix0 ON perfdata_service (host_name);
CREATE INDEX perfdata_service_ix1 ON perfdata_service (last_perfdata_raw);
CREATE INDEX perfdata_service_ix2 ON perfdata_service (raw_delete_policy_name);

--
-- Tables to hold the raw perfdata for the services:
--

CREATE TABLE perfdata_service_raw (
	id SERIAL NOT NULL PRIMARY KEY,
	host_name VARCHAR(75) NOT NULL,
	service_description VARCHAR(75) NOT NULL,
	FOREIGN KEY (host_name, service_description) 
		REFERENCES perfdata_service (host_name, service_description) 
		ON DELETE RESTRICT,
	ctime TIMESTAMP NOT NULL,
	txt_data VARCHAR(255),
	nagios_status INT2 DEFAULT NULL
);

CREATE INDEX perfdata_service_raw_ix0 ON perfdata_service_raw (host_name, service_description);
CREATE INDEX perfdata_service_raw_ix1 ON perfdata_service_raw (ctime);
CREATE INDEX perfdata_service_raw_ix2 ON perfdata_service_raw (host_name, service_description, ctime);

ALTER TABLE perfdata_service 
	ADD FOREIGN KEY (last_perfdata_raw) 
	REFERENCES perfdata_service_raw(id) 
	ON DELETE SET NULL;

--
-- Table to hold the metrics:
--

CREATE TABLE perfdata_service_metric (
	metric_id INT NOT NULL,
	UNIQUE(metric_id),
	host_name VARCHAR(75) NOT NULL,
	service_description VARCHAR(75) NOT NULL,
	metric VARCHAR(75) NOT NULL,
	PRIMARY KEY (host_name, service_description, metric),
	FOREIGN KEY (host_name, service_description) 
		REFERENCES perfdata_service (host_name, service_description) 
		ON DELETE RESTRICT,
	unit VARCHAR(10),
	notes TEXT,
	value_max FLOAT,
	value_min FLOAT,
	last_perfdata_bin INTEGER,
	bin_delete_policy INTEGER DEFAULT NULL,
	bin_delete_policy_name VARCHAR(20) DEFAULT NULL,
	bin_delete_policy_type VARCHAR(10) CHECK (bin_delete_policy_type IN('template','user','host')) NOT NULL, 
	FOREIGN KEY (bin_delete_policy_name)
		REFERENCES perfdata_delete_policy (policy_name)
		ON DELETE SET NULL,
	is_deleted BOOL DEFAULT FALSE
	);
	
CREATE INDEX perfdata_service_metric_ix0 ON perfdata_service_metric (host_name, service_description);
CREATE UNIQUE INDEX perfdata_service_metric_ix1 ON perfdata_service_metric (last_perfdata_bin);

--
-- Table to hold the binary perfdata
--

CREATE TABLE perfdata_service_bin (
	id SERIAL NOT NULL PRIMARY KEY,
	host_name VARCHAR(75) NOT NULL,
	service_description VARCHAR(75) NOT NULL,
	metric VARCHAR(75) NOT NULL,
	FOREIGN KEY (host_name, service_description) 
		REFERENCES perfdata_service (host_name, service_description) 
		ON DELETE RESTRICT,
	FOREIGN KEY (host_name, service_description, metric) 
		REFERENCES perfdata_service_metric (host_name, service_description, metric) 
		ON DELETE RESTRICT,
	ctime TIMESTAMP NOT NULL,
	value FLOAT,
	warn FLOAT,
	critical FLOAT,
	state INT2
	);
	
CREATE INDEX perfdata_service_bin_ix0 ON perfdata_service_bin (host_name, service_description);
CREATE INDEX perfdata_service_bin_ix1 ON perfdata_service_bin (host_name, service_description, metric);
CREATE INDEX perfdata_service_bin_ix2 ON perfdata_service_bin (host_name, service_description, metric, ctime);

ALTER TABLE perfdata_service_metric 
	ADD FOREIGN KEY (last_perfdata_bin) 
	REFERENCES perfdata_service_bin(id) 
	ON DELETE SET NULL;

--
-- Binary Summary Data
-- Delete policy and frequency in Hours.
--

CREATE TABLE perfdata_bin_summary_del_policy (
	policy_name VARCHAR(20) NOT NULL PRIMARY KEY,
	delete_policy INT DEFAULT NULL
	);

CREATE TABLE perfdata_bin_summary (
	frequency INT NOT NULL,
	PRIMARY KEY (frequency),
	notes TEXT NOT NULL DEFAULT '',
	is_deleted BOOL NOT NULL DEFAULT FALSE,
	default_delete INT NOT NULL DEFAULT 1000
	);

CREATE TABLE perfdata_bin_summary_header (
	metric_id INT NOT NULL,
	frequency INT NOT NULL,
	PRIMARY KEY (metric_id, frequency),
	FOREIGN KEY (metric_id) 
		REFERENCES perfdata_service_metric (metric_id) 
		ON DELETE RESTRICT,
	FOREIGN KEY (frequency) 
		REFERENCES perfdata_bin_summary (frequency) 
		ON DELETE RESTRICT,
	last_data INTEGER,
	bin_delete_policy INTEGER DEFAULT NULL,
	bin_delete_policy_name VARCHAR(20) DEFAULT NULL,
	bin_delete_policy_type VARCHAR(10) CHECK (bin_delete_policy_type IN('template','user')) NOT NULL,
	is_deleted BOOL DEFAULT FALSE
	);
	
CREATE INDEX perfdata_bin_summary_header_ix0 ON perfdata_bin_summary_header (frequency);

CREATE TABLE perfdata_bin_summary_data (
	metric_id INT NOT NULL,
	frequency INT NOT NULL,
	ctime TIMESTAMP NOT NULL,
	PRIMARY KEY (metric_id, frequency, ctime),
	FOREIGN KEY (metric_id, frequency) 
		REFERENCES perfdata_bin_summary_header (metric_id, frequency)
		ON DELETE RESTRICT,
	val_count INTEGER NOT NULL DEFAULT 0,
	sum_val FLOAT8 NOT NULL,
	sum_square_val FLOAT8 NOT NULL,
	max_val FLOAT8 NOT NULL,
	min_val FLOAT8 NOT NULL,
	max_warn FLOAT8, -- NOT NULL,
	min_warn FLOAT8, -- NOT NULL,
	max_critical FLOAT8, -- NOT NULL,
	min_critical FLOAT8 -- NOT NULL
	);

--
-- Table to hold the system state:
--

CREATE TABLE perfdata_state (
	domain VARCHAR(10) NOT NULL PRIMARY KEY,
	file_name TEXT NOT NULL DEFAULT '',
	line_in_file INTEGER NOT NULL DEFAULT 0,
	last_line_hash VARCHAR(20) NOT NULL DEFAULT '',
	pid INTEGER NOT NULL DEFAULT 0,
	perfparse_version VARCHAR(10) NOT NULL DEFAULT '0'
	);

CREATE TABLE perfdata_registry (
		host VARCHAR(75) DEFAULT NULL,
		rkey VARCHAR(80) NOT NULL,
		PRIMARY KEY (host, rkey),
		rvalue TEXT,
		ctime TIMESTAMP NOT NULL
);

--
-- Table to hold users perferances,
-- for instance filters used in CGI
--

CREATE TABLE perfdata_prefs (
	user_name VARCHAR(20) NOT NULL PRIMARY KEY,
	group_name VARCHAR(75),
	FOREIGN KEY (group_name) 
		REFERENCES perfdata_host_group (group_name) 
		ON DELETE SET NULL
	);

CREATE INDEX perfdata_prefs_ix0 ON perfdata_prefs (group_name);

--
-- Stored graphs:
--

CREATE TABLE perfdata_graphs (
	id SERIAL PRIMARY KEY NOT NULL,
	group_name VARCHAR(75),
	FOREIGN KEY (group_name) 
		REFERENCES perfdata_host_group (group_name) 
		ON DELETE SET NULL,
	title VARCHAR(80),
	descript TEXT,
	url TEXT,
	refresh_time INT NOT NULL DEFAULT 0
	);

CREATE INDEX perfdata_graphs_ix0 ON perfdata_graphs (group_name);

--
-- Remembers User Settings
--

CREATE TABLE perfdata_users (
	user_name VARCHAR(20) NOT NULL,
	PRIMARY KEY (user_name),
	password VARCHAR(20),
	group_name VARCHAR(75),
		--REFERENCES perfdata_host_group (group_name) 
		--ON DELETE SET NULL,
	raw_show_normal BOOL NOT NULL DEFAULT TRUE,
	raw_show_warning BOOL NOT NULL DEFAULT TRUE,
	raw_show_critical BOOL NOT NULL DEFAULT TRUE,
	raw_show_unknown BOOL NOT NULL DEFAULT TRUE,
	raw_refresh_time INT NOT NULL DEFAULT 0
	);


